% -------------------------------------------------------------------------
%       Synthesis of compliant mechanisms with selective compliance  
%                           An advanced procedure
%
%         Stephanie Kirmse, Lucio Flavio Campanile, Alexander Hasse
%                         Mechanism and Machine Theory
% -------------------------------------------------------------------------

% *************************** Disclaimer *********************************
% The authors reserve all rights but do not guarantee that the code is
% free from errors. In addition, they are in no case liable for any event 
% caused by the use of the program.  
% ************************************************************************

% the numbers in brackets correspond to equations from the paper
clear all
%%%%%%%%%%%%%%%%%%%%%%Input of the required parameters%%%%%%%%%%%%%%%%%%%%%
%parametrization
bb=40;                  %width of design space [mm]
h=30;                   %height of design space [mm]
numb=12;                %number of elements on width
numh=9;                 %number of elements on height
%material and dimensions of beam elements
width=10;               %width beam elements [mm]
thick=2;                %thickness of beam elements [mm]
E=210000;               %Young's modulus [N/mm^2]
volumefraction=0.7;     %distributable volume based on initial volume
%boundary conditions
%%%%%%%%%%%%%%%%%%
%node arrangement%
%       3 4      %
%       1 2      %
%%%%%%%%%%%%%%%%%%
fixeddofs=1:39;         %fixed degrees of freedom
phi=[1 0 1 0];          %desired deformation mode as row vector 
modedofs=[352 353 388 389]; %master-DoFs
%parameters for optimization
numM=3;                 %number of modes for stabilization according to (22)
v=0.005;                %nu (26)
kappa=0;                %kappa (27)
numIte=1000;            %number of iterations
mu=1000;                %upper limit to the primary stiffness mu

%%%%%%%%%%%%%%%%%%%%%%%%%calculations%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
phi=(1/sqrt(phi*phi')).*phi';          %normalization of phi according to(5)
%allocation matrix and node coordinates
[ZUO,PuKo]=BeamTruss(h,bb,numb,numh);
beams=length(ZUO);
%calculation of stiffness matrix
[Standardmatrix]=Standardmatrices(ZUO,PuKo,width,thick,E);
%subdivision of degrees of freedom
alldofs = 1:length(PuKo)*3;             %all degrees of freedom
freedofs = setdiff(alldofs,fixeddofs);  %free degrees of freedom
modefree  = setdiff(freedofs,modedofs); %slave-DoFs
%limits according to (24)
xmin=0.00000001*ones(beams,1);
xmax=1*ones(beams,1);
%beam degrees of freedom matrix
BEAMS=zeros(beams,6);
for i=1:beams
    BEAMS(i,:)=[(3*(ZUO(i,2))-2) (3*(ZUO(i,2))-2)+1 (3*(ZUO(i,2))-2)+2 ...
        (3*(ZUO(i,3))-2) (3*(ZUO(i,3))-2)+1 (3*(ZUO(i,3))-2)+2];
end
iK = reshape(kron(BEAMS,ones(6,1))',36*beams,1);
jK = reshape(kron(BEAMS,ones(1,6))',36*beams,1);
standard=cell2mat(Standardmatrix);
d3=zeros(6,6,beams);
for i=1:beams
    d3(:,:,i)=Standardmatrix{1,i};
end
%random start vector 
x0=random('unif',xmin,xmax);
fvalg=zeros(numIte,1);

%%%%%%%%%%%%%%%%%%%%%global iteration procedure%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for i=1:numIte
    %%%%%%%%%%%%%calculation objective function and restrictions%%%%%%%%%%%
    %objective function (20), restrictions (21) and (22)
    [A,f]=Displacement_distribution(x0,alldofs,...
    phi,modedofs,numM,BEAMS,modefree,iK,jK,standard,d3,beams);
    b(1)=mu;
    b(2:length(A(:,1)))=0;
    %volume_constraint (23)
    A(length(A(:,1))+1,:)=ones(1,beams);
    b(length(A(:,1)))=length(xmax)*volumefraction;
    %%%%%%%%%%%sub-problem 2: updating the design variables%%%%%%%%%%%%%%%%
    options = optimoptions(@linprog,'Algorithm','dual-simplex',...
        'MaxIter',3000,'Preprocess','none','Tolcon',1e-3,'Display','none');
    [x,fval,exitflag,output]=linprog(f,A,b,[],[],xmin,xmax,...
        options);
    if exitflag==-2 || exitflag==0
        x=x0;
        disp(['no solution can be found with these input parameters and'...
        'this starting vector']);
        break;
    end
    %convergence history
    fvalg(i)=fval;
    
    %%%%%%%%%%%stabilization of the global iteration procedure%%%%%%%%%%%%%
    %stabilization (27)
    x0=kappa*x0+(1-kappa)*x;
    % stabilization (25)
    xmax=x0+v;
    xmax(xmax>1)=1;
    xmin=x0-v;
    xmin(xmin<0.00000001)=0.00000001;
    disp(['Iteration: ' num2str(i)]);
end

%%%%%%%%plot the result and eigenvalue analysis according (31)%%%%%%%%%%%%%
[chi,lambda]=Plotdef(x,ZUO,PuKo,freedofs,alldofs,h,bb,modedofs,thick,...
    beams,modefree,standard,iK,jK);
%eigenvalue analysis
[lambdasort,ind]=sort(diag(lambda));        %eigenvalues
chisort=chi(:,ind);                         %eigenvectors
S=lambdasort(2)/lambdasort(1);              %selelctivity according (32)